(* queue.ml: persistent queue using two stacks of incoming and
   outgoing elements. *)

type 'a queue = {               (* immutable fields *)
  incoming : 'a list;           (* incoming for enqueue *)
  outgoing : 'a list;           (* outgoing for dequeue *)
};;

(* An empty queue record *)
let empty = {incoming=[]; outgoing=[]};;

(* Return a queue with x pushed into the incoming stack *)
let enqueue q x =
  {q with incoming=(x::q.incoming)}
;;

(* Return a queue with one element removed; pops an element from
   outgoing stack if not empty; otherwise reverses incoming list, pops
   an element and then assigns remainder to outgoing leaving incoming
   empty. raises and exception on both incoming/outgoing empty. *)
let dequeue q =
  match q with
  | {outgoing=[]; incoming=[]} -> failwith "empty queue"
  | {outgoing=front::rest} ->     {q with outgoing=rest}
  | _ ->                          (* outgoing empty, incoming present *)
     let revin = List.rev q.incoming in
     {incoming=[]; outgoing=(List.tl revin)}
;;

(* Returns a pair of the front element of a queue and a potentially
   re-arranged queue.  Re-arrangement is done like in dequeue if the
   outgoing stack is empty. Exception on both incoming/outgoing
   empty. *)
let front q =
  match q with
  | {outgoing=[]; incoming=[]} -> failwith "empty queue"
  | {outgoing=front::rest} ->     (front,q)
  | _ ->                          (* outgoing empty, incoming present *)
     let revin = List.rev q.incoming in
     let front = List.hd revin in
     let newq = {incoming=[]; outgoing=revin} in
     (front,newq)
;;
